/*
 * @(#)FastCharset.java        0.0        09/07/29
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 * Contributors: Ulf Zibis, Germany, Ulf.Zibis @ CoSoCo.de
 */

package sun.nio.cs;

import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CoderResult;

/**
 * Class for charsets using lazy loaded mappings.
 *
 * @author Ulf Zibis, Cologne <Ulf.Zibis @ CoSoCo.de>
 */
public class FastCharset extends Charset implements HistoricallyNamedCharset {

    private final String pkgName;
    final String mappingClsName; // package private for ExternalCharsetProvider
    private transient Decoder decoder;
    private transient Encoder encoder;

    FastCharset(String className, String canonicalName, String... aliases) {
        super(canonicalName, aliases);
        pkgName = getClass().getPackage().getName();
        mappingClsName = className;
    }
    FastCharset(String packageName, String className, String canonicalName, String[] aliases) {
        super(canonicalName, aliases);
        pkgName = packageName;
        mappingClsName = className;
    }

    public String historicalName() {
        return mapping().historicalName;
    }

    public boolean contains(Charset cs) {
        return cs.name().equals(name()) ||
                mapping().isASCII && cs.name().equals("US-ASCII");
    }

    public Decoder newDecoder() {
        return decoder().clone();
    }

    public Encoder newEncoder() {
        return encoder().clone();
    }

    Decoder decoder() {
        if (decoder == null)
            decoder = mapping().newDecoder(this);
        return decoder;
    }

    Encoder encoder() {
        if (encoder == null)
            encoder = decoder().mapping.newEncoder(this);
        return encoder;
    }

    final <M extends Mapping> M mapping() {
        Error error = new InternalError("Can't instantiate class "+pkgName+'.'+mappingClsName);
        try {
            return ((Class<M>) Class.forName(
                pkgName+'.'+mappingClsName)).newInstance();
        } catch (ClassNotFoundException ex) {
            throw new NoClassDefFoundError(ex.getMessage());
//        } catch (ReflectiveOperationException ex) {
//            Error error = new InternalError("Can't instantiate class "+pkgName+'.'+mappingClsName);
//            error.initCause(ex);
//            throw error;
        } catch (InstantiationException ex) {
            error.initCause(ex);
        } catch (IllegalAccessException ex) {
            error.initCause(ex);
        }
        throw error;
    }


    protected static abstract class Decoder<M extends Mapping>
            extends CharsetDecoder implements Cloneable, ArrayDecoder {

        protected final M mapping;
        private ByteBuffer srcBuf;
        private CharBuffer dstBuf;

        protected Decoder(FastCharset cs, M mapping,
                float averageCharsPerByte, float maxCharsPerByte) {
            super(cs, averageCharsPerByte, maxCharsPerByte);
            this.mapping = mapping;
        }

        protected Decoder<M> clone() {
            try {
                return (Decoder<M>)super.clone();
            } catch (CloneNotSupportedException ex) {
                return null;
            }
        }

        protected CoderResult decodeLoop(ByteBuffer src, CharBuffer dst) {
            srcBuf = src;
            if (src.hasArray() && dst.hasArray()) {
                dstBuf = dst;
                int sp = src.arrayOffset() + src.position();
                int dp = dst.arrayOffset() + dst.position();
                return decodeArrayLoop(src.array(), sp, sp + src.remaining(),
                        dst.array(), dp, dp + dst.remaining());
            } else
                return decodeBufferLoop(src, dst);
        }

        protected abstract CoderResult decodeArrayLoop(
                byte[] sa, int sp, int sl, char[] da, int dp, int dl);

        protected abstract CoderResult decodeBufferLoop(
                ByteBuffer src, CharBuffer dst);


        protected final CoderResult malformed1(int sp) {
            updateSrcPosition(sp);
            return CoderResult.MALFORMED_1;
        }

        protected final CoderResult malformed2(int sp) {
            updateSrcPosition(sp);
            return CoderResult.MALFORMED_2;
        }

        protected final CoderResult unmappable1(int sp) {
            updateSrcPosition(sp);
            return CoderResult.UNMAPPABLE_1;
        }

        protected final CoderResult unmappable2(int sp) {
            updateSrcPosition(sp);
            return CoderResult.UNMAPPABLE_2;
        }

        protected final CoderResult underflow(int sp) {
            updateSrcPosition(sp);
            return CoderResult.UNDERFLOW;
        }

        protected final CoderResult overflow(int sp) {
            updateSrcPosition(sp);
            return CoderResult.OVERFLOW;
        }

        protected final CoderResult malformed1(int sp, int dp) {
            updateDstPosition(dp);
            return malformed1(sp);
        }

        protected final CoderResult malformed2(int sp, int dp) {
            updateDstPosition(dp);
            return malformed2(sp);
        }

        protected final CoderResult unmappable1(int sp, int dp) {
            updateDstPosition(dp);
            return unmappable1(sp);
        }

        protected final CoderResult unmappable2(int sp, int dp) {
            updateDstPosition(dp);
            return unmappable2(sp);
        }

        protected final CoderResult underflow(int sp, int dp) {
            updateDstPosition(dp);
            return underflow(sp);
        }

        protected final CoderResult overflow(int sp, int dp) {
            updateDstPosition(dp);
            return overflow(sp);
        }

        final void updateSrcPosition(int sp) {
            if (sp < 0)
                srcBuf.position(srcBuf.position() + sp);
            else
                srcBuf.position(sp - srcBuf.arrayOffset());
        }

        final void updateDstPosition(int dp) {
            dstBuf.position(dp - dstBuf.arrayOffset());
        }
    }


    protected static abstract class Encoder<M extends Mapping>
            extends CharsetEncoder implements Cloneable, ArrayEncoder {

        private CharBuffer srcBuf;
        private int srcOff;
        private ByteBuffer dstBuf;
        private int dstOff;
        
        protected Encoder(FastCharset cs,
                   float averageBytesPerChar, float maxBytesPerChar) {
            super(cs, averageBytesPerChar, maxBytesPerChar);
        }

        protected Encoder<M> clone() {
            try {
                return (Encoder<M>)super.clone();
            } catch (CloneNotSupportedException ex) {
                return null;
            }
        }

        protected CoderResult encodeLoop(CharBuffer src, ByteBuffer dst) {
            srcBuf = src;
            if (src.hasArray() && dst.hasArray()) {
                dstBuf = dst;
                srcOff = src.arrayOffset();
                int sp = srcOff + src.position();
                dstOff = dst.arrayOffset();
                int dp = dstOff + dst.position();
                return encodeArrayLoop(src.array(), sp, sp + src.remaining(),
                        dst.array(), dp, dp + dst.remaining());
            } else
                return encodeBufferLoop(src, dst);
        }

        protected abstract CoderResult encodeArrayLoop(
                char[] sa, int sp, int sl, byte[] da, int dp, int dl);

        protected abstract CoderResult encodeBufferLoop(
                CharBuffer src, ByteBuffer dst);


        protected final CoderResult malformed1(int sp) {
            updateSrcPosition(sp);
            return CoderResult.MALFORMED_1;
        }

        protected final CoderResult malformed2(int sp) {
            updateSrcPosition(sp);
            return CoderResult.MALFORMED_2;
        }

        protected final CoderResult unmappable1(int sp) {
            updateSrcPosition(sp);
            return CoderResult.UNMAPPABLE_1;
        }

        protected final CoderResult unmappable2(int sp) {
            updateSrcPosition(sp);
            return CoderResult.UNMAPPABLE_2;
        }

        protected final CoderResult underflow(int sp) {
            updateSrcPosition(sp);
            return CoderResult.UNDERFLOW;
        }

        protected final CoderResult overflow(int sp) {
            updateSrcPosition(sp);
            return CoderResult.OVERFLOW;
        }

        protected final CoderResult malformed1(int sp, int dp) {
            updateDstPosition(dp);
            return malformed1(sp);
        }

        protected final CoderResult malformed2(int sp, int dp) {
            updateDstPosition(dp);
            return malformed2(sp);
        }

        protected final CoderResult unmappable1(int sp, int dp) {
            updateDstPosition(dp);
            return unmappable1(sp);
        }

        protected final CoderResult unmappable2(int sp, int dp) {
            updateDstPosition(dp);
            return unmappable2(sp);
        }

        protected final CoderResult underflow(int sp, int dp) {
            updateDstPosition(dp);
            return underflow(sp);
        }

        protected final CoderResult overflow(int sp, int dp) {
            updateDstPosition(dp);
            return overflow(sp);
        }

        final void updateSrcPosition(int sp) {
            if (sp < 0)
                srcBuf.position(srcBuf.position() + sp);
            else
                srcBuf.position(sp - srcOff);
        }

        final void updateDstPosition(int dp) {
            dstBuf.position(dp - dstOff);
        }
    }
}
