/*
 * @(#)SingleByte.java        0.0        09/07/29
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 * Contributors: Ulf Zibis, Germany, Ulf.Zibis @ CoSoCo.de
 */

package sun.nio.cs;

import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.CoderResult;
import java.util.Arrays;
import static sun.nio.cs.Mapping.UNMAPPABLE_ENCODING;
import static sun.nio.cs.Mapping.REPLACE_CHAR;
import static sun.nio.cs.Mapping.REPLACE_CHARS;
// TODO: use constants from CoderResult:
// import static java.nio.charset.CoderResult.*;

class SingleByte {

    final static class Decoder extends FastCharset.Decoder<SingleByteMapping> {

        private final char[] b2c;

        Decoder(FastCharset cs, SingleByteMapping mapping) {
            super(cs, mapping, 1.0f, 1.0f);
            b2c = mapping.b2c;
        }

        protected CoderResult decodeArrayLoop(
                byte[] sa, int sp, int sl, char[] da, int dp, int dl) {
            boolean srcLongerThanDst = (sl - sp) > (dl - dp);
            if (srcLongerThanDst)
                sl = sp + (dl - dp);
            while (sp < sl) {
                char c = decode(sa[sp++]);
                if (c == REPLACE_CHAR)
                    return unmappable1(sp-1, dp);
                da[dp++] = c;
            }
            return srcLongerThanDst ? overflow(sp-0, dp) : underflow(sp-0, dp);
        }

        protected CoderResult decodeBufferLoop(ByteBuffer src, CharBuffer dst) {
            while (src.hasRemaining()) {
                char c = decode(src.get());
                if (c == REPLACE_CHAR)
                    return unmappable1(1);
                if (!dst.hasRemaining())
                    return overflow(1);
                dst.put(c);
            }
            return CoderResult.UNDERFLOW;
        }

        private final char decode(int b) {
            return b2c[b + 128];
        }

        private char repl = REPLACE_CHAR; // TODO: better refer to default of Charset class?
        protected void implReplaceWith(String newReplacement) {
            repl = newReplacement.charAt(0);
        }

        public int decode(byte[] src, int sp, int len, char[] dst) {
            if (len > dst.length)
                len = dst.length;
            int dp = 0;
            while (dp < len) {
                char c = decode(src[sp++]);
                dst[dp++] = (c == REPLACE_CHAR) ? repl : c;
            }
            return dp;
        }
    }


    final static class Encoder extends FastCharset.Encoder<SingleByteMapping> {

        final char[] c2bIndex;
        final char[] c2b;

        Encoder(FastCharset cs, SingleByteMapping mapping) {
            super(cs, 1.0f, 1.0f);
            c2bIndex = REPLACE_CHARS.clone();
            c2b = new char[mapping.c2bLength];
            // init the c2b and c2bIndex tables from b2c.
            final char[] nrBytes = mapping.nrBytes;
            final char[] c2bChars = mapping.c2bChars;
            final char[] c2bBytes = mapping.c2bBytes;
            char[] b2c = mapping.b2c;
            if (nrBytes != null) {
                b2c = b2c.clone();
                for (char c : nrBytes)
                    b2c[((byte)c)+0x80] = REPLACE_CHAR;
            }
            Arrays.fill(c2b, REPLACE_CHAR);
            char off = 0;
            for (int i = 0; i < b2c.length; i++)
                off = initEncoding(b2c[i], (char)((i+0x80) & 0xff), off);
            if (c2bChars != null)
                // c-->b nr entries
                for (int i=0; i<c2bChars.length; i++)
                    off = initEncoding(c2bChars[i], c2bBytes[i], off);
        }

        private char initEncoding(char c, char b, char off) {
            if (c != REPLACE_CHAR) {
                int index = c >> 8;
                if (c2bIndex[index] == UNMAPPABLE_ENCODING) {
                    c2bIndex[index] = off;
                    off += 0x100;
                }
                c2b[c2bIndex[index]+(c & 0xff)] = b;
            }
            return off;
        }

        public boolean canEncode(char c) {
            return encode(c) != REPLACE_CHAR;
        }

        public boolean isLegalReplacement(byte[] repl) {
            return ((repl.length == 1 && repl[0] == (byte)'?') ||
                    super.isLegalReplacement(repl));
        }

        protected CoderResult encodeArrayLoop(
                char[] sa, int sp, int sl, byte[] da, int dp, int dl) {
            boolean srcLongerThanDst = (sl - sp) > (dl - dp);
            if (srcLongerThanDst)
                sl = sp + (dl - dp);
            while (sp < sl) {
                char c = sa[sp++];
                int b = encode(c);
                if (b == UNMAPPABLE_ENCODING) {
                    if (c >= Character.MIN_HIGH_SURROGATE) {
                        if (c <= Character.MAX_HIGH_SURROGATE) { // = isHigh
                            if (sp == sl && !srcLongerThanDst)
                                return underflow(sp-1, dp);
                            if (Character.isLowSurrogate(sa[sp++]))
                                return unmappable2(sp-2, dp);
                            return malformed1(sp-2, dp);
                        } else if (c <= Character.MAX_LOW_SURROGATE) // = isLow
                            return malformed1(sp-1, dp);
                    }
                    return unmappable1(sp-1, dp);
                }
                da[dp++] = (byte)b;
            }
            return srcLongerThanDst ? overflow(sp-0, dp) : underflow(sp-0, dp);
        }

        protected CoderResult encodeBufferLoop(CharBuffer src, ByteBuffer dst) {
            while (src.hasRemaining()) {
                char c = src.get();
                int b = encode(c);
                if (b == UNMAPPABLE_ENCODING) {
                    if (c >= Character.MIN_HIGH_SURROGATE) {
                        if (c <= Character.MAX_HIGH_SURROGATE) { // = isHigh
                            if (!src.hasRemaining())
                                return underflow(1);
                            if (Character.isLowSurrogate(src.get()))
                                return unmappable2(2);
                            return malformed1(2);
                        } else if (c <= Character.MAX_LOW_SURROGATE) // = isLow
                            return malformed1(1);
                    }
                    return unmappable1(1);
               }
                if (!dst.hasRemaining())
                    return overflow(1);
                dst.put((byte)b);
            }
            return CoderResult.UNDERFLOW;
        }

        private final int encode(char ch) {
            char index = c2bIndex[ch >> 8];
            if (index == UNMAPPABLE_ENCODING)
                return UNMAPPABLE_ENCODING;
            return c2b[index + (ch & 0xff)];
        }

        private byte repl = (byte)'?'; // TODO: better refer to default of Charset class?
        protected void implReplaceWith(byte[] newReplacement) {
            repl = newReplacement[0]; // TODO: what's about replacements with size > 1 ?
        }

        public int encode(char[] src, int sp, int len, byte[] dst) {
            int dp = 0;
            int sl = sp + Math.min(len, dst.length);
            while (sp < sl) {
                char c = src[sp++];
                int b = encode(c);
                if (b == UNMAPPABLE_ENCODING) {
                    b = repl;
                    if (Character.isHighSurrogate(c) && sp < sl &&
                            Character.isLowSurrogate(src[sp])) {
                        sp++;
                        if (len > dst.length) {
                            sl++;
                            len--;
                        }
                    }
                }
                dst[dp++] = (byte)b;
            }
            return dp;
        }
    }
}
