/*
 * @(#)UTF_16.java        1.21 05/11/17
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 * Contributors: Ulf Zibis, Germany, Ulf.Zibis @ CoSoCo.de
 */

package sun.nio.cs;

import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CoderResult;

class UTF_16 extends Unicode {

    UTF_16(String canonicalName, String... aliases) {
        super(canonicalName, aliases);
    }

    public String historicalName() {
        return "UTF-16";
    }

    public Decoder newDecoder() {
        return new Decoder(this, NONE);
    }

    public Encoder newEncoder() {
        return new Encoder(this, BIG, true);
    }


    /**
     * Base class for different flavors of UTF-16 decoders
     */
    static class Decoder extends CharsetDecoder {

        private final int expectedByteOrder;
        private int currentByteOrder;
        private int defaultByteOrder = BIG;

        Decoder(Charset cs, int bo) {
            super(cs, 0.5f, 1.0f);
            expectedByteOrder = currentByteOrder = bo;
        }

        Decoder(Charset cs, int bo, int defaultBO) {
            this(cs, bo);
            defaultByteOrder = defaultBO;
        }

        private char decode(int b1, int b2) {
            if (currentByteOrder == BIG)
                return (char)((b1 << 8) | b2);
            else
                return (char)((b2 << 8) | b1);
        }

        protected CoderResult decodeLoop(ByteBuffer src, CharBuffer dst) {
            int mark = src.position();

            try {
                while (src.remaining() > 1) {
                    int b1 = src.get() & 0xff;
                    int b2 = src.get() & 0xff;

                    // Byte Order Mark interpretation
                    if (currentByteOrder == NONE) {
                        char c = (char)((b1 << 8) | b2);
                        if (c == BYTE_ORDER_MARK) {
                            currentByteOrder = BIG;
                            mark += 2;
                            continue;
                        } else if (c == REVERSED_MARK) {
                            currentByteOrder = LITTLE;
                            mark += 2;
                            continue;
                        } else {
                            currentByteOrder = defaultByteOrder;
                            // FALL THROUGH to process b1, b2 normally
                        }
                    }

                    char c = decode(b1, b2);

                    if (c == REVERSED_MARK) {
                        // A reversed BOM cannot occur within middle of stream
                        return CoderResult.malformedForLength(2);
                    }

                    // Surrogates
                    if (Surrogate.is(c)) {
                        if (Surrogate.isHigh(c)) {
                            if (src.remaining() < 2)
                                return CoderResult.UNDERFLOW;
                            char c2 = decode(src.get() & 0xff, src.get() & 0xff);
                            if (!Surrogate.isLow(c2))
                                return CoderResult.malformedForLength(4);
                            if (dst.remaining() < 2)
                                return CoderResult.OVERFLOW;
                            mark += 4;
                            dst.put(c);
                            dst.put(c2);
                            continue;
                        }
                        // Unpaired low surrogate
                        return CoderResult.malformedForLength(2);
                    }

                    if (!dst.hasRemaining())
                        return CoderResult.OVERFLOW;
                    mark += 2;
                    dst.put(c);

                }
                return CoderResult.UNDERFLOW;

            } finally {
                src.position(mark);
            }
        }

        protected void implReset() {
            currentByteOrder = expectedByteOrder;
        }
    }


    /**
     * Base class for different flavors of UTF-16 encoders
     */
    static class Encoder extends CharsetEncoder {

        private int byteOrder;      /* Byte order in use */
        private boolean usesMark;   /* Write an initial BOM */
        private boolean needsMark;

        Encoder(Charset cs, int bo, boolean m) {
            super(cs, 2.0f,
                  // Four bytes max if you need a BOM
                  m ? 4.0f : 2.0f,
                  // Replacement depends upon byte order
                  ((bo == BIG)
                   ? new byte[] { (byte)0xff, (byte)0xfd }
                   : new byte[] { (byte)0xfd, (byte)0xff }));
            usesMark = needsMark = m;
            byteOrder = bo;
        }

        private void put(char c, ByteBuffer dst) {
            if (byteOrder == BIG) {
                dst.put((byte)(c >> 8));
                dst.put((byte)(c & 0xff));
            } else {
                dst.put((byte)(c & 0xff));
                dst.put((byte)(c >> 8));
            }
        }

        private final Surrogate.Parser sgp = new Surrogate.Parser();

        protected CoderResult encodeLoop(CharBuffer src, ByteBuffer dst) {
            int mark = src.position();

            if (needsMark) {
                if (dst.remaining() < 2)
                    return CoderResult.OVERFLOW;
                put(BYTE_ORDER_MARK, dst);
                needsMark = false;
            }

            try {
                while (src.hasRemaining()) {
                    char c = src.get();
                    if (!Surrogate.is(c)) {
                        if (dst.remaining() < 2)
                            return CoderResult.OVERFLOW;
                        mark++;
                        put(c, dst);
                        continue;
                    }
                    int d = sgp.parse(c, src);
                    if (d < 0)
                        return sgp.error();
                    if (dst.remaining() < 4)
                        return CoderResult.OVERFLOW;
                    mark += 2;
                    put(Surrogate.high(d), dst);
                    put(Surrogate.low(d), dst);
                }
                return CoderResult.UNDERFLOW;
            } finally {
                src.position(mark);
            }
        }

        protected void implReset() {
            needsMark = usesMark;
        }

        public boolean canEncode(char c) {
            return ! Surrogate.is(c);
        }
    }
}
