/*
 * @(#)UTF_32.java        1.1        05/11/09
 *
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 * Contributors: Ulf Zibis, Germany, Ulf.Zibis @ CoSoCo.de
 */

package sun.nio.cs;

import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CoderResult;

class UTF_32 extends Unicode {

    private static final int BOM_BIG = BYTE_ORDER_MARK;
    private static final int BOM_LITTLE = REVERSED_MARK << 0x10;

    UTF_32(String canonicalName, String... aliases) {
        super(canonicalName, aliases);
    }

    public String historicalName() {
        return "UTF-32";
    }

    public Decoder newDecoder() {
        return new Decoder(this, NONE);
    }

    public Encoder newEncoder() {
        return new Encoder(this, BIG, false);
    }


/**
 * Base class for different flavors of UTF-32 decoders
 */
    static class Decoder extends CharsetDecoder {
        private int currentBO;
        private int expectedBO;

        Decoder(Charset cs, int bo) {
            super(cs, 0.25f, 1.0f);
            this.expectedBO = bo;
            this.currentBO = NONE;
        }

        private int getCP(ByteBuffer src) {
            return (currentBO == BIG)
                    ? (((src.get() & 0xff) << 24) |
                    ((src.get() & 0xff) << 16) |
                    ((src.get() & 0xff) << 8) |
                    (src.get() & 0xff))
                    : ((src.get() & 0xff) |
                    ((src.get() & 0xff) << 8) |
                    ((src.get() & 0xff) << 16) |
                    ((src.get() & 0xff) << 24));
        }

        protected CoderResult decodeLoop(ByteBuffer src, CharBuffer dst) {
            if (src.remaining() < 4)
                return CoderResult.UNDERFLOW;
            int mark = src.position();
            int cp;
            try {
                if (currentBO == NONE) {
                    cp = ((src.get() & 0xff) << 24) |
                         ((src.get() & 0xff) << 16) |
                         ((src.get() & 0xff) <<  8) |
                         (src.get() & 0xff);
                    if (cp == BOM_BIG && expectedBO != LITTLE) {
                        currentBO = BIG;
                        mark += 4;
                    } else if (cp == BOM_LITTLE && expectedBO != BIG) {
                        currentBO = LITTLE;
                        mark += 4;
                    } else {
                        if (expectedBO == NONE)
                            currentBO = BIG;
                        else
                            currentBO = expectedBO;
                        src.position(mark);
                    }
                }
                while (src.remaining() > 3) {
                    cp = getCP(src);
                    if (cp < 0 || cp > Surrogate.UCS4_MAX)
                        return CoderResult.malformedForLength(4);
                    if (cp < Surrogate.UCS4_MIN) {
                        if (!dst.hasRemaining())
                            return CoderResult.OVERFLOW;
                        mark += 4;
                        dst.put((char)cp);
                    } else {
                        if (dst.remaining() < 2)
                            return CoderResult.OVERFLOW;
                        mark += 4;
                        dst.put(Surrogate.high(cp));
                        dst.put(Surrogate.low(cp));
                    }
                }
                return CoderResult.UNDERFLOW;
            } finally {
                src.position(mark);
            }
        }

        protected void implReset() {
            currentBO = NONE;
        }
    }


/**
 * Base class for different flavors of UTF-32 encoders
 */
    static class Encoder extends CharsetEncoder {
        private boolean doBOM = false;
        private boolean doneBOM = true;
        private int byteOrder;

        Encoder(Charset cs, int byteOrder, boolean doBOM) {
            super(cs, 4.0f,
                    doBOM?8.0f:4.0f,
                    (byteOrder==BIG)? new byte[]{(byte)0, (byte)0, (byte)0xff, (byte)0xfd}
                                    : new byte[]{(byte)0xfd, (byte)0xff, (byte)0, (byte)0});
            this.byteOrder = byteOrder;
            this.doBOM = doBOM;
            this.doneBOM = !doBOM;
        }

        private void put(int cp, ByteBuffer dst) {
            if (byteOrder==BIG) {
                dst.put((byte)(cp >> 24));
                dst.put((byte)(cp >> 16));
                dst.put((byte)(cp >> 8));
                dst.put((byte)cp);
            } else {
                dst.put((byte)cp);
                dst.put((byte)(cp >>  8));
                dst.put((byte)(cp >> 16));
                dst.put((byte)(cp >> 24));
            }
        }

        protected CoderResult encodeLoop(CharBuffer src, ByteBuffer dst) {
            int mark = src.position();
            if (!doneBOM) {
                if (dst.remaining() < 4)
                    return CoderResult.OVERFLOW;
                put(BOM_BIG, dst);
                doneBOM = true;
            }
            try {
                while (src.hasRemaining()) {
                    char c = src.get();
                    if (Surrogate.isHigh(c)) {
                        if (!src.hasRemaining())
                            return CoderResult.UNDERFLOW;
                        char low = src.get();
                        if (Surrogate.isLow(low)) {
                            if (dst.remaining() < 4)
                                return CoderResult.OVERFLOW;
                            mark += 2;
                            put(Surrogate.toUCS4(c, low), dst);
                        } else {
                            return CoderResult.malformedForLength(1);
                        }
                    } else if (Surrogate.isLow(c)) {
                        return CoderResult.malformedForLength(1);
                    } else {
                        if (dst.remaining() < 4)
                            return CoderResult.OVERFLOW;
                        mark++;
                        put(c, dst);
                    }
                }
                return CoderResult.UNDERFLOW;
            } finally {
                src.position(mark);
            }
        }

        protected void implReset() {
            doneBOM = !doBOM;
        }
    }
}
